import java.nio.charset.StandardCharsets._
import java.nio.file.{Files, Paths}
import io.shiftleft.passes.{CpgPass, DiffGraph}
import io.shiftleft.codepropertygraph.Cpg
import io.shiftleft.dataflowengine.language.{Flow, FlowPrettyPrinterExt}
import java.util.Calendar
import java.util.Date
import java.text.SimpleDateFormat
import java.text.DateFormat

val sdf = new SimpleDateFormat("yyyy-MM-dd'T'HH:mm:ss'Z'")

case class DataFlowElement(sequence : Int,
	file : String,
	lineNumber : Int,
    columnNumber : Int,
    text : String
)
implicit val dataFlowW = upickle.default.macroW[DataFlowElement]

case class StaticDetails(file : String,
	parameter : String,
	dataFlow : List[DataFlowElement]
)
implicit val staticDetailsW = upickle.default.macroW[StaticDetails]

case class Mapping(mappingType : String, value : String)
implicit val mappingW = upickle.default.macroW[Mapping]

case class Finding(nativeId : String,
	severity : String,
	nativeSeverity : String,
    mappings: List[Mapping],
	summary : String,
	scannerRecommendation : String,
	staticDetails : StaticDetails
)
implicit val findingW = upickle.default.macroW[Finding]


case class ThreadFix(id : Long  ,
	created : String,
	exported : String,
	collectionType : String,
	source : String,
    executiveSummary : String,
	findings : List[Finding])
implicit val threadFixW = upickle.default.macroW[ThreadFix]


def creatingCpg(payload: String, payloadType : String) : Boolean = {
    println("[+] Verify if CPG exists")
    if(!workspace.baseCpgExists(payload)) {

        payloadType match {

            case "JAR" | "WAR" | "EAR" =>

                printf("[+] Creating CPG and SP for %s\n", payload)
                createCpgAndSp(payload)

                println("[+] Verify if CPG was created successfully")
                if(!workspace.baseCpgExists(payload)) {
                    printf("[+] Failed to create CPG for %s\n", payload)
                    return false
                }
            case "CPG" =>

                println("[+] Creating CPG for " + payload)
                loadCpg(payload)
                addOverlay("tagging")
                addOverlay("securityprofile")

                println("[+] Verify if CPG was created successfully")
                if(!workspace.baseCpgExists(payload)) {
                    printf("[+] Failed to create CPG for %s\n", payload)
                    return false
                }

            case _ =>
                println("[+] Unrecognized payload type specified")
        }

    } else {
        println("[+] CPG already exists. Proceed to loadCpg")
    }
    return true
}

def loadingCpg(payload: String, payloadType : String) : Boolean = {
    println("[+] Load if CPG exists")
    if(workspace.baseCpgExists(payload)) {

        printf("[+] Loading pre-existing CPG for %s\n", payload)
        loadCpg(payload)

        if(workspace.recordExists(payload)) {
            printf("[+] CPG successfully loaded for %s\n", payload)
        } else {
            printf("[+] Failed to load CPG for %s\n", payload)
            return false
        }
    } else {
        printf("[+] Attempting to load non-existent CPG for %s\n", payload)
        return false
    }
    return true
}

def getSeverity(score : Float) : String = {
    score match {
        case s if (s >= 0 && s < 3) => "LOW"
        case s if (s >= 3 && s < 6) => "MEDIUM"
        case s if (s >= 6 && s <= 9) => "HIGH"
        case _ => "UNKNOWN"
    }
}

def getMappingValueFromLink(link : String) = {
    link.split("/").toList.filter(_.contains(".html")).head.split("\\.").filter(!_.contains("html")).toList.head
}

def getThreadFixExport(payloadIdentity : Long, cpg: io.shiftleft.codepropertygraph.Cpg) = {

	val findings = cpg.finding.l.map { f =>

		val pointData = f.ioflows.map(_.primaryFlow.points.map { point =>
	    	FlowPrettyPrinterExt.pointToData(point.elem.asInstanceOf[nodes.TrackingPoint])
		}).flatten.zipWithIndex

	    val dataFlow = pointData.map(pd => DataFlowElement(pd._2,
            pd._1.fileOption.getOrElse("NA"),
            pd._1.lineNumberOption.getOrElse(0).asInstanceOf[Int],
            0,
            pd._1.methodNameOption.getOrElse("NA") + "(" + pd._1.trackedSymbol + ")")
        )

	    val staticDetails = StaticDetails(f.filename,
            pointData.head._1.trackedSymbol,
            dataFlow)


	    Finding(f.id.toString,
			getSeverity(f.score),
			getSeverity(f.score),
            List(Mapping("CWE",getMappingValueFromLink(f.cweLink))),
			f.category + ":" + f.title,
			f.description,
			staticDetails)
	}

	ThreadFix(payloadIdentity,
        sdf.format(new Date()),
        sdf.format(new Date()),
        "SAST",
        "SHIFTLEFT_IO",
        "Analysis Results",
        findings)

}

@doc("")
@main def execute(payloadIdentity : Long,
				payload: String,
                payloadType: String,
                outFile: String) : Boolean = {

    if(creatingCpg(payload,payloadType) && loadingCpg(payload,payloadType)) {

    	println("Writing to OutFile : " + outFile)
        val writer = new java.io.PrintWriter(new java.io.File(outFile))
        writer.write(upickle.default.write(getThreadFixExport(payloadIdentity,cpg), indent=2))
       	writer.close()

        printf("[+] Saving results to %s\n", outFile)

    } else {
        printf("[+] Failed to execute THREADFIX script for CPG/ARTIFACT of Payload %s\n", payload)
    }
    return true

}